#include <cstdio>
#include <csignal>
#include <fcntl.h>
#include <unistd.h>
#include <vector>

/*
    
    This program gives mp3 to stdout.
    Don't invoke this multiple times.
    Instead, copy the output.

*/

#include "config.hh"

int main(void)
{
    int pip[2];
    int monpid;
    char *Buf;
    pipe(pip);
    
    chdir("/");
    
    vector<const char *> argv;

    int nulfd = open("/dev/null", O_RDWR);
    dup2(nulfd, 2);
    
    monpid = fork();
    if(monpid < 0)return -1;

    if(!monpid)
    {
        dup2(nulfd, 0);
        dup2(pip[1], 1);
        close(pip[0]);
        close(pip[1]);
        
        string esdmon = Configuration.GetString("esdmonbinaryname");
        string mixmon = Configuration.GetString("mixmonbinaryname");
        string monhost = Configuration.GetString("monitorhost");
        string filter = Configuration.GetString("filter");
        
        if(filter.size())
        {
            argv.push_back(mixmon.c_str());
            if(monhost.size())
            {
                Buf = static_cast<char *>(malloc(monhost.size() + 3));
                sprintf(Buf, "-H%s", monhost.c_str());
                argv.push_back(Buf);
            }
            
            Buf = static_cast<char *>(malloc(filter.size() + 3));
            sprintf(Buf, "-f%s", filter.c_str());
            argv.push_back(Buf);
            
            Buf = static_cast<char *>(malloc(128));
            char *d = Buf;
            *d++ = '-';
            *d++ = 'F';
            *d = 0;
            if(!Configuration.GetBool("stereo"))
            {
                *d++ = 'm';
                *d = 0;
            }
            if(Configuration.GetInteger("inputherz") != 44100)
            {
                sprintf(d, "r%u", Configuration.GetInteger("inputherz"));
            }
            argv.push_back(Buf);
            argv.push_back(NULL);
            execvp(mixmon.c_str(), const_cast<char **>(&argv[0]));
        }
        else
        {
            argv.push_back(esdmon.c_str());
            if(monhost.size())
            {
                argv.push_back("-s");
                argv.push_back(monhost.c_str());
            }
            if(!Configuration.GetBool("stereo"))
                argv.push_back("-m");
            if(Configuration.GetInteger("inputherz") != 44100)
            {
                Buf = static_cast<char *>(malloc(64));
                sprintf(Buf, "%u", Configuration.GetInteger("inputherz"));
                argv.push_back("-r");
                argv.push_back(Buf);
            }
            argv.push_back(NULL);
            execvp(esdmon.c_str(), const_cast<char **>(&argv[0]));
        }
        _exit(-1);
    }
    
    sleep(1); // ease strace
    
    dup2(pip[0], 0);
    close(pip[0]);
    close(pip[1]);
    
    /* Skip esdmon message (ends with '\n'). We have safety counter
     * anyway, for the what-if -case if message doesn't appear.
     * Note the safety counter limit must be divisible by 4,
     * so that we don't mess stereo channels or endianess.
     */
    for(unsigned skipcount=0; skipcount<80; ++skipcount)
    {
        char c;
        if(read(0, &c, 1) < 1 || c=='\n')break;
    }
    
    string lame = Configuration.GetString("lamebinaryname");
    
    argv.push_back(lame.c_str());
    argv.push_back("-rxd");

    if(!Configuration.GetBool("stereo"))
    {
        argv.push_back("-m");
        argv.push_back("m");
    }

    argv.push_back("-h");

    string Q = Configuration.GetString("quality");
    string V = Configuration.GetString("vbrquality");
    string avkb = Configuration.GetString("avgkbrate");
    string mikb = Configuration.GetString("minkbrate");
    string makb = Configuration.GetString("maxkbrate");
    
    if(Q.size() && Q != "5") { argv.push_back("-q"); argv.push_back(Q.c_str()); }
    if(V.size() && V != "4") { argv.push_back("-V"); argv.push_back(V.c_str()); }
    if(avkb.size()) { argv.push_back("--abr"); argv.push_back(avkb.c_str()); }
    if(mikb.size()) { argv.push_back("-b"); argv.push_back(mikb.c_str()); }
    if(makb.size()) { argv.push_back("-B"); argv.push_back(makb.c_str()); }
    
    unsigned inputhz = Configuration.GetInteger("inputherz");
    unsigned outputhz = Configuration.GetInteger("streamherz");
    
    //if(outputhz != inputhz)
    //{
        Buf = static_cast<char *>(malloc(64));
        sprintf(Buf, "%.3f", outputhz / 1000.0);
        argv.push_back("--resample");
        argv.push_back(Buf);
    //}
    if(inputhz != 44100)
    {
        Buf = static_cast<char *>(malloc(64));
        sprintf(Buf, "%.3f", inputhz / 1000.0);
        argv.push_back("-s");
        argv.push_back(Buf);
    }
    
    argv.push_back("-");
    argv.push_back("-");

    argv.push_back(NULL);
    execvp(lame.c_str(), const_cast<char **>(&argv[0]));
    
    printf("Couldn't start '%s'\n", lame.c_str());
    close(0);
    kill(monpid, SIGTERM);
    
    exit(-1);
}
