#include <cstdio>
#include <csignal>
#include <sys/wait.h>
#include <unistd.h>
#include <vector>

/*

    This program does the tcp stuff.
    This program also starts and stops the encoder if necessary.
    This is quite the center.

*/

#include "config.hh"
#include "logfun.h"
#include "radiofun.hh"
#include "serverfun.h"

static unsigned volatile AbortMe;
static void Alarm(int p)
{
    if(p == SIGCHLD && AbortMe)
        kill(getpid(), SIGABRT);
}

static void Listen(vector<int> &Plugs)
{
    string ports = Configuration.GetString("listen");
    
    unsigned a=0;
    while(a < ports.size())
    {
        unsigned p = 0;
        while(a<ports.size() && isdigit(ports[a]))
            p = p*10 + (ports[a++]-'0');
        
        if(p)
        {
            int fd = MakeListenerSocket(p);
            if(fd >= 0)Plugs.push_back(fd);
        }
        
        while(a < ports.size() && !isdigit(ports[a]))++a;
    }
}

int main(void)
{
    vector<int> Plugs;
    
    Listen(Plugs);
    
    vector<httpclient> clients;
    
    signal(SIGCHLD, Alarm);
    signal(SIGABRT, Alarm);
    signal(SIGPIPE, SIG_IGN);
    
    for(;;)
    {
        fd_set fds;
        FD_ZERO(&fds);
        int hiest = 0;
        for(unsigned a=0; a<Plugs.size(); ++a)
        {
            FD_SET(Plugs[a], &fds);
            if(Plugs[a] > hiest)hiest = Plugs[a];
        }
        if(Server.fd >= 0)
        {
            FD_SET(Server.fd, &fds);
            if(Server.fd > hiest)hiest = Server.fd;
        }
        
        AbortMe = 1;
        /* Wait until something happens. */
        if(select(hiest+1, &fds, NULL, NULL, NULL) <= 0)
            FD_ZERO(&fds);
        AbortMe = 0;
        
        for(unsigned a=0; a<Plugs.size(); ++a)
            if(FD_ISSET(Plugs[a], &fds))
            {
                httpclient clnt(Plugs[a]);
                if(clnt.fd >= 0) /* Got a client here */
                {
                    clnt.Exec();
                    clients.push_back(clnt);
                }
            }
        
        if(Server.fd >= 0 && FD_ISSET(Server.fd, &fds))
        {
            Server.forward();
        }
        
        int pid, status;
        while((pid = waitpid(-1, &status, WNOHANG)) > 0)
        {
            if(WIFEXITED(status))
            {
                status = WEXITSTATUS(status);
                bool found = false;
                
                unsigned a=0;
                for(a=0; a<clients.size(); )
                {
                    if(clients[a].pid != pid) { ++a; continue; }
                    found = true;
                    if(clients[a].GotClient(status))
                    {
                        Server.addtarget(clients[a]);
                    }
                    clients.erase(clients.begin() + a);
                }
                
                #if 0
                if(Verbose)
                {
                    DumpTime();
                    fprintf(stderr, "Pid %d (", pid);
                    if(found)
                        fprintf(stderr, "http");
                    else
                        fprintf(stderr, "encoder");
                    
                    fprintf(stderr, ") exited (");
                    if(status == 55)
                        fprintf(stderr, "and now streaming begins");
                    else
                        fprintf(stderr, "%d", status);
                    
                    fprintf(stderr, ")\n");
                }
                #endif
                
                if(!found)
                {
                    DumpTime();
                    fprintf(stderr, "We are completely idle again.\n"
                                    "---------------------\n");
                }
            }
        }
    }
    return 0;
}
