#include <stdio.h>
#include <limits.h>
#include <math.h>

#include "chrtab.h"

unsigned short Lookup[rmax][gmax][bmax];

unsigned char OldAttr = -1;
unsigned char TextAttr = -1;

int Use8bit = 1;

#if 0
/* rgb_to_hsv:
 *  Converts an RGB value into the HSV colorspace.
 */
void rgb_to_hsv(int r, int g, int b, float *h, float *s, float *v)
{
   float min, max, delta, rc, gc, bc;

   rc = (float)r / 63.0;
   gc = (float)g / 63.0;
   bc = (float)b / 63.0;
   
   max=rc; if(gc>max)max=gc; if(bc>max)max=bc;
   min=rc; if(gc<min)min=gc; if(bc<min)min=bc;
   delta = max - min;
   *v = max;

   if (max != 0.0)
      *s = delta / max;
   else
      *s = 0.0;

   if (*s == 0.0) {
      *h = 0.0; 
   }
   else {
      if (rc == max)
	 *h = (gc - bc) / delta;
      else if (gc == max)
	 *h = 2 + (bc - rc) / delta;
      else if (bc == max)
	 *h = 4 + (rc - gc) / delta;

      *h *= 60.0;
      if (*h < 0)
	 *h += 360.0;
    }
   *h /= 360.0;
}
#endif

#if MIRC<2
static const char *ColorDiff(unsigned char old, unsigned char new)
{
#if MIRC
	static char Buff[8];
	static const char mirk[16] = {1, 2, 3,10, 5, 6, 7,15,
	                             14,12, 9,11, 4,13, 8,0};
	
	if(old==new)return "";
	
	sprintf(Buff, "%d", mirk[new&15]);
	if(old==255 || ((old>>4) != (new>>4)))
		sprintf(strchr(Buff, 0), ",%d", mirk[new>>4]);
	
#else
	static char Buff[16];
	char *be = Buff;
	
	/* Todo:
	 * If the background color of old and new is the same,
	 * and is writing only spaces, there is no need to
	 * change the colour - it will still look the same.
	 */
	
    if(TextAttr == OldAttr)
    {
    	Buff[0] = 0;
    	return Buff;
    }
	be[0] = 27;
	be[1] = '[';
	be += 2;

#if !BROKEN_TERM
	if(new != 7)
#endif
    {
    	static char Swap[] = "04261537";
    	
    	int pp=0;
    	int lostblink = (old&0x80) > (new&0x80);
    	int lostinten = (old&0x08) > (new&0x08);
    	
    	if(lostblink || lostinten)
    	{
    	#if BROKEN_TERM
    		*be++ = '0'; pp=1;
    		/* Ensure all attributes are reset */
    		old = (new&0x77) ^ 0x77;
    	#else
    		if(lostblink && !lostinten)
    		{
    			*be++ = '2'; *be++ = '5'; pp = 1;
    			old &= 0x7F;
    		}
    		else if(lostinten && !lostblink)
    		{
	    		*be++ = '2'; pp=1;
	    		old &= 0xF7;
    		}
    		else
    		{
	    		*be++ = '0'; pp=1;
    			old = 7;
    		}
    	#endif
    	}    	
    	
    	if((new&0x08) && !(old&0x08)){if(pp)*be++ = ';'; *be++ = '1';pp=1;}
    	if((new&0x80) && !(old&0x80)){if(pp)*be++ = ';'; *be++ = '5';pp=1;}
       	
       	if((new&0x70) != (old&0x70))
       	{
   	   		if(pp)*be++ = ';';
   	   		*be++ = '4';
   	   		*be++ = Swap[(new>>4)&7];
       		pp=1;
       	}
       	
      	if((new&7) != (old&7))
      	{
       		if(pp)*be++ = ';';
       		*be++ = '3';
       		*be++ = Swap[new&7];
       	}
       	*be = 0;
    }
    
    be[0] = 'm';
    be[1] = 0;
#endif
    
    return Buff;
}

int FlushSetAttr()
{
	const char *s = ColorDiff(OldAttr, TextAttr);
	printf("%s", s);
    
	OldAttr = TextAttr;
	
	return *s;
}
#endif

const char DefR[16] = {0, 0, 0, 0,42,42,42, 42, 21, 21,21,21,63,63,63,63};
const char DefG[16] = {0, 0,42,42, 0, 0,21, 42, 21, 21,63,63,21,21,63,63};
const char DefB[16] = {0,42, 0,42, 0,42, 0, 42, 21, 63,21,63,21,63,21,63};
	
void MkTab(void)
{
	#define palmax 63
	
	#define MRKNUM 5
	
	static char Merkit[MRKNUM] = " ";
	static char Bri[MRKNUM] = {0, 2*8, 4*8, 6*8, 8*8};
		
	#define brimax (8*8)
	
	int r, r1[16][16][MRKNUM];
	int g, g1[16][16][MRKNUM];
	int b, b1[16][16][MRKNUM];

	int backlimit = Use8bit ? 16 : 8;

	int fore, back, a;

	for(back=0; back<backlimit; back++)
		for(fore=0; fore<16; fore++)
			for(a=0; a<MRKNUM; a++)
			{
				r1[back][fore][a] =
					(Bri[a] * DefR[fore] + (brimax-Bri[a]) * DefR[back]) / brimax;

				g1[back][fore][a] =
					(Bri[a] * DefG[fore] + (brimax-Bri[a]) * DefG[back]) / brimax;

				b1[back][fore][a] =
					(Bri[a] * DefB[fore] + (brimax-Bri[a]) * DefB[back]) / brimax;
			}
		
	#undef brimax

	for(r=0; r<rmax; r++)
		for(g=0; g<gmax; g++)
			for(b=0; b<bmax; b++)
			{
				int BB=0,BF=0,BA=0,Sc=INT_MAX;
				
				int r2 = r * (palmax-1) / (rmax-1);
				int g2 = g * (palmax-1) / (gmax-1);
				int b2 = b * (palmax-1) / (bmax-1);
				
				#if 0
				
				float H,S,V;				
				rgb_to_hsv(r2,g2,b2, &H,&S,&V);
				
				#endif

				for(back=0; back<backlimit; back++)
					for(fore=0; fore<16; fore++)
						for(a=0; a<MRKNUM; a++)
						{
							#if 1
							
							int R = (r1[back][fore][a] - r2);
							int G = (g1[back][fore][a] - g2);
							int B = (b1[back][fore][a] - b2);

							int Bla = R*R + G*G + B*B;
							
							#else
							
							float h,s,v;
							float dh,ds,dv;
							int Bla;
							rgb_to_hsv(
								r1[back][fore][a],
								g1[back][fore][a],
								b1[back][fore][a],
								&h,&s,&v);
								
							dh = fabs(h-H);	if(dh > 0.5)dh = 1.0-dh;
							ds = fabs(s-S);
							dv = fabs(v-V);
							Bla = 50000 * (4*dh + 1*ds + 2*dv);
							
							#endif

							if(Bla<Sc || (!Bla && a<BA))
								BB=back, BF=fore, BA=a, Sc=Bla;
						}

				Lookup[r][g][b] =
					(BB<<12) |
					(BF<<8) |
					(unsigned char)BA[Merkit];
			}
			
	#undef palmax
	#undef MRKNUM
}
